import asyncio
import concurrent.futures
import functools
import logging


def cleanup_bear(bear,
                 running_tasks,
                 event_loop):
    """
    Cleans up state of an ongoing run for a bear.

    - If the given bear has no running tasks left, it removes the bear from
      the ``running_tasks`` dict.
    - Checks whether there are any remaining tasks, and quits the event loop
      accordingly if none are left.

    :param bear:
        The bear to clean up state for.
    :param running_tasks:
        The dict of running-tasks.
    :param event_loop:
        The event-loop tasks are scheduled on.
    """
    if not running_tasks[bear]:
        del running_tasks[bear]

    if not running_tasks:
        event_loop.stop()


def schedule_bears(bears,
                   result_callback,
                   event_loop,
                   running_tasks,
                   executor):
    """
    Schedules the tasks of bears to the given executor and runs them on the
    given event loop.

    :param bears:
        A list of bear instances to be scheduled onto the process pool.
    :param result_callback:
        A callback function which is called when results are available.
    :param event_loop:
        The asyncio event loop to schedule bear tasks on.
    :param running_tasks:
        Tasks that are already scheduled, organized in a dict with
        bear instances as keys and asyncio-coroutines as values containing
        their scheduled tasks.
    :param executor:
        The executor to which the bear tasks are scheduled.
    """
    for bear in bears:
        tasks = {
            event_loop.run_in_executor(
                executor, bear.execute_task, bear_args, bear_kwargs)
            for bear_args, bear_kwargs in bear.generate_tasks()}

        running_tasks[bear] = tasks

        for task in tasks:
            task.add_done_callback(functools.partial(
                finish_task, bear, result_callback,
                running_tasks, event_loop, executor))

        logging.debug('Scheduled {!r} (tasks: {})'.format(bear,
                                                          len(tasks)))

        if not tasks:
            # We need to recheck our runtime if something is left to
            # process, as when no tasks were offloaded the event-loop could
            # hang up otherwise.
            cleanup_bear(bear, running_tasks, event_loop)


def finish_task(bear,
                result_callback,
                running_tasks,
                event_loop,
                executor,
                task):
    """
    The callback for when a task of a bear completes. It is responsible for
    checking if the bear completed its execution and the handling of the
    result generated by the task.

    :param bear:
        The bear that the task belongs to.
    :param result_callback:
        A callback function which is called when results are available.
    :param running_tasks:
        Dictionary that keeps track of the remaining tasks of each bear.
    :param event_loop:
        The ``asyncio`` event loop bear-tasks are scheduled on.
    :param executor:
        The executor to which the bear tasks are scheduled.
    :param task:
        The task that completed.
    """
    try:
        results = task.result()
    except Exception as ex:
        # FIXME Try to display only the relevant traceback of the bear if error
        # FIXME occurred there, not the complete event-loop traceback.
        logging.error('An exception was thrown during bear execution.',
                      exc_info=ex)

        results = None
    finally:
        running_tasks[bear].remove(task)
        cleanup_bear(bear, running_tasks, event_loop)

    if results is not None:
        for result in results:
            try:
                # FIXME Long operations on the result-callback could block the
                # FIXME   scheduler significantly. It should be possible to
                # FIXME   schedule new Python Threads on the given event_loop
                # FIXME   and process the callback there.
                result_callback(result)
            except Exception as ex:
                # FIXME Try to display only the relevant traceback of the
                # FIXME result handler if error occurred there, not the
                # FIXME complete event-loop traceback.
                logging.error(
                    'An exception was thrown during result-handling.',
                    exc_info=ex)


def run(bears, result_callback):
    """
    Runs a coala session.

    :param bears:
        The bear instances to run.
    :param result_callback:
        A callback function which is called when results are available. Must
        have following signature::

            def result_callback(result):
                pass
    """
    # FIXME Allow to pass different executors nicely, for example to execute
    # FIXME   coala with less cores, or to schedule jobs on distributed systems
    # FIXME   (for example Mesos).

    # Set up event loop and executor.
    event_loop = asyncio.SelectorEventLoop()
    executor = concurrent.futures.ProcessPoolExecutor()

    # Let's go.
    schedule_bears(bears, result_callback, event_loop, {}, executor)
    try:
        event_loop.run_forever()
    finally:
        event_loop.close()



import logging
import re

from coala_utils.string_processing.StringConverter import StringConverter
from coala_utils.string_processing import (unescape, convert_to_raw,
                                           position_is_escaped,
                                           unescaped_rstrip)


class LineParser:

    def __init__(self,
                 key_value_delimiters=('=',),
                 comment_separators=('#',),
                 key_delimiters=(',', ' '),
                 section_name_surroundings=None,
                 section_override_delimiters=('.',),
                 key_value_append_delimiters=('+=',)):
        """
        Creates a new line parser. Please note that no delimiter or separator
        may be an "o" or you may encounter undefined behaviour with the
        escapes.

        :param key_value_delimiters:        Delimiters that delimit a key from
                                            a value.
        :param comment_separators:          Used to initiate a comment.
        :param key_delimiters:              Delimiters between several keys.
        :param section_name_surroundings:   Dictionary, e.g. {"[", "]"} means a
                                            section name is surrounded by [].
                                            If None, {"[": "]"} is used as
                                            default.
        :param section_override_delimiters: Delimiter for a section override.
                                            E.g. "." would mean that
                                            section.key is a possible key that
                                            puts the key into the section
                                            "section" despite of the current
                                            section.
        :param key_value_append_delimiters: Delimiters to separate key and
                                            value in setting arguments where
                                            settings are being appended.
        """
        section_name_surroundings = (
            {'[': ']'} if section_name_surroundings is None
            else section_name_surroundings)

        self.key_value_delimiters = key_value_delimiters
        self.key_value_append_delimiters = key_value_append_delimiters
        self.comment_separators = comment_separators
        self.key_delimiters = key_delimiters
        self.section_name_surroundings = section_name_surroundings
        self.section_override_delimiters = section_override_delimiters

    def parse(self, line):
        """
        Note that every value in the returned tuple *besides the value* is
        unescaped. This is so since the value is meant to be put into a Setting
        later thus the escapes may be needed there.

        :param line: The line to parse.
        :return:     section_name (empty string if it's no section name),
                     [(section_override, key), ...], value, comment
        """
        logging.warning('The parse method of LineParser is deprecated and will'
                        ' be removed. Please use `_parse` which has a new '
                        'return type, a tuple containing 5 values instead of '
                        '4. Refer to the method documentation for further '
                        'information.')
        section_name, key_tuples, value, _, comment = self._parse(line)
        return section_name, key_tuples, value, comment

    def _parse(self, line):
        """
        Note that every value in the returned tuple *besides the value* is
        unescaped. This is so since the value is meant to be put into a Setting
        later thus the escapes may be needed there.

        :param line: The line to parse.
        :return:     section_name (empty string if it's no section name),
                     [(section_override, key), ...], value, to_append (True if
                     append delimiter is found else False), comment
        """
        for separator in self.comment_separators:
            if (re.match('[^ ]' + separator, line)
                    or re.match(separator + '[^ ]', line)):
                logging.warning('This comment does not have whitespace' +
                                ' before or after ' + separator + ' in: ' +
                                repr(line.replace('\n', '')) + '. If you ' +
                                'didn\'t mean to make a comment, use a ' +
                                'backslash for escaping.')

        line, comment = self.__separate_by_first_occurrence(
            line,
            self.comment_separators)
        comment = unescape(comment)
        if line == '':
            return '', [], '', False, comment

        section_name = unescape(self.__get_section_name(line))
        if section_name != '':
            return section_name, [], '', False, comment

        # Escapes in value might be needed by the bears
        append = True
        keys, value = self.__extract_keys_and_value(
            line, self.key_value_append_delimiters)
        if not value:
            keys, value = self.__extract_keys_and_value(
                line, self.key_value_delimiters, True)
            append = False

        # Add all the delimiters that stored as tuples
        all_delimiters = self.key_value_delimiters
        all_delimiters += self.key_value_append_delimiters
        all_delimiters += self.key_delimiters
        all_delimiters += self.comment_separators
        all_delimiters += self.section_override_delimiters
        all_delimiters = ''.join(all_delimiters)

        # Add all keys and values in section_name_surroundings, which is
        # stored as a dict
        all_delimiters += ''.join(self.section_name_surroundings.keys())
        all_delimiters += ''.join(self.section_name_surroundings.values())

        value = convert_to_raw(value, all_delimiters)

        key_tuples = []
        for key in keys:
            key = convert_to_raw(key, all_delimiters)
            section, key = self.__separate_by_first_occurrence(
                key,
                self.section_override_delimiters,
                True,
                True)
            key_tuples.append((unescape(section), unescape(key)))

        return '', key_tuples, value, append, comment

    @staticmethod
    def __separate_by_first_occurrence(string,
                                       delimiters,
                                       strip_delim=False,
                                       return_second_part_nonempty=False):
        """
        Separates a string by the first of all given delimiters. Any whitespace
        characters will be stripped away from the parts.

        :param string:                      The string to separate.
        :param delimiters:                  The delimiters.
        :param strip_delim:                 Strips the delimiter from the
                                            result if true.
        :param return_second_part_nonempty: If no delimiter is found and this
                                            is true the contents of the string
                                            will be returned in the second part
                                            of the tuple instead of the first
                                            one.
        :return:                            (first_part, second_part)
        """
        temp_string = string.replace('\\\\', 'oo')
        i = temp_string.find('\\')
        while i != -1:
            temp_string = temp_string[:i] + 'oo' + temp_string[i+2:]
            i = temp_string.find('\\', i+2)

        delim_pos = len(string)
        used_delim = ''
        for delim in delimiters:
            pos = temp_string.find(delim)
            if 0 <= pos < delim_pos:
                delim_pos = pos
                used_delim = delim

        if return_second_part_nonempty and delim_pos == len(string):
            return '', string.strip(' \n')

        first_part = string[:delim_pos]
        second_part = string[delim_pos + (
            len(used_delim) if strip_delim else 0):]

        if not position_is_escaped(second_part, len(second_part) - 1):
            first_part = unescaped_rstrip(first_part)
            second_part = unescaped_rstrip(second_part)

        return (first_part.lstrip().rstrip('\n'),
                second_part.lstrip().rstrip('\n'))

    def __get_section_name(self, line):
        for begin, end in self.section_name_surroundings.items():
            if (line[0:len(begin)] == begin and
                    line[len(line) - len(end):len(line)] == end):
                return line[len(begin):len(line) - len(end)].strip(' \n')

        return ''

    def __extract_keys_and_value(self,
                                 line,
                                 delimiters,
                                 return_second_part_nonempty=False):
        """
        This method extracts the keys and values from the give string by
        splitting them based on the delimiters provided.

        :param line:                        The input string.
        :param delimiters:                  A list of delimiters to split the
                                            strings on.
        :param return_second_part_nonempty: If no delimiter is found and this
                                            is true the contents of the string
                                            will be returned as value
        :return:                            The parsed keys and values from a
                                            line.
        """
        key_part, value = self.__separate_by_first_occurrence(
            line,
            delimiters,
            True,
            return_second_part_nonempty)
        keys = list(StringConverter(
            key_part,
            list_delimiters=self.key_delimiters).__iter__(
            remove_backslashes=False))

        return keys, value



from coalib.bears.Bear import Bear
from coalib.bears.BEAR_KIND import BEAR_KIND


class GlobalBear(Bear):
    """
    A GlobalBear analyzes semantic facts across several files.

    The results of a GlobalBear will be presented grouped by the origin Bear.
    Therefore Results spanning across multiple files are allowed and will be
    handled correctly.

    If you are inspecting a single file at a time, you should consider
    using a LocalBear.
    """

    def __init__(self,
                 file_dict,
                 section,
                 message_queue,
                 timeout=0):
        """
        Constructs a new GlobalBear.

        :param file_dict: The dictionary of {filename: file contents}.

        See :class:`coalib.bears.Bear` for other parameters.
        """
        Bear.__init__(self, section, message_queue, timeout)
        self.file_dict = file_dict

    @staticmethod
    def kind():
        return BEAR_KIND.GLOBAL

    def run(self,
            *args,
            dependency_results=None,
            **kwargs):
        """
        Handles all files in file_dict.

        :param dependency_results: The dictionary of {bear name:
                                   result list}.
        :return: A list of Result type.

        See :class:`coalib.bears.Bear` for `run` method description.
        """
        raise NotImplementedError(
            'This function has to be implemented for a runnable bear.')



from coalib.misc.Enum import enum

RESULT_SEVERITY = enum('INFO', 'NORMAL', 'MAJOR')
RESULT_SEVERITY.__str__ = lambda x: RESULT_SEVERITY.reverse.get(x, 'NORMAL')
RESULT_SEVERITY_COLORS = {RESULT_SEVERITY.INFO: 'green',
                          RESULT_SEVERITY.NORMAL: 'yellow',
                          RESULT_SEVERITY.MAJOR: 'red'}



import logging
from coalib.misc.Enum import enum

LOG_LEVEL = enum(DEBUG=logging.DEBUG,
                 INFO=logging.INFO,
                 WARNING=logging.WARNING,
                 ERROR=logging.ERROR)
LOG_LEVEL_COLORS = {LOG_LEVEL.ERROR: 'red',
                    LOG_LEVEL.WARNING: 'yellow',
                    LOG_LEVEL.INFO: 'blue',
                    LOG_LEVEL.DEBUG: 'green'}



from coala_utils.decorators import (
    enforce_signature, generate_ordering, generate_repr)


@generate_repr('line', 'column')
@generate_ordering('line', 'column')
class TextPosition:

    @enforce_signature
    def __init__(self, line: (int, None)=None, column: (int, None)=None):
        """
        Creates a new TextPosition object that represents the position inside
        a string with line/column numbers.

        :param line:        The line in file or None, the first line is 1.
        :param column:      The column indicating the character. The first one
                            in a line is 1.
        :raises TypeError:  Raised when line or columns are no integers.
        :raises ValueError: Raised when a column is set but line is None.
        """
        if line is None and column is not None:
            raise ValueError('A column can only be set if a line is set.')

        self._line = line
        self._column = column

    @property
    def line(self):
        return self._line

    @property
    def column(self):
        return self._column

    def __le__(self, other):
        """
        Test whether ``self`` is behind or equals the other
        ``TextPosition``.

        If the column in a ``TextPosition`` is ``None``, consider
        whole line. If the line in a ``TextPosition`` is ``None``,
        consider whole file.

        :param other: ``TextPosition`` to compare with.
        :return:      Whether this ``TextPosition`` is behind the other
                      one or the same.
        """
        if self.line is None or other.line is None:
            return True
        if self.line == other.line:
            return (True
                    if self.column is None or
                    other.column is None
                    else
                    self.column <= other.column)
        else:
            return self.line < other.line

    def __ge__(self, other):
        """
        Test whether ``self`` is ahead of or equals the
        other ``TextPosition``.

        If the column in a ``TextPosition`` is ``None``, consider
        whole line. If the line in a ``TextPosition`` is ``None``,
        consider whole file.

        :param other: ``TextPosition`` to compare with.
        :return:      Whether this ``TextPosition`` is ahead of the other
                      one or the same.
        """
        if self.line is None or other.line is None:
            return True
        if self.line == other.line:
            return (True
                    if self.column is None or
                    other.column is None
                    else
                    self.column >= other.column)
        else:
            return self.line > other.line



import re
import sys
from importlib import import_module
from inspect import getmembers
from pkgutil import iter_modules
from types import ModuleType

from .base import aspectbase
from .meta import aspectclass, aspectTypeError
from .taste import Taste, TasteError

# already import Root here to make it available in submodules that define
# `Root.subaspect`s, which get imported in wrapper `aspectsModule.__init__`
# before this module gets replaced with that wrapper in sys.modules
from .root import Root


__all__ = ['Root', 'Taste', 'TasteError',
           'aspectclass', 'aspectbase', 'aspectTypeError']


class aspectsModule(ModuleType):
    """
    A special module wrapper for ``coalib.bearlib.aspects``, allowing
    case-insensitive aspect lookup by name via direct indexing :)

    >>> import coalib.bearlib.aspects

    >>> coalib.bearlib.aspects['Metadata']
    <aspectclass 'Root.Metadata'>

    >>> coalib.bearlib.aspects['commitmessage']
    <aspectclass 'Root.Metadata.CommitMessage'>

    >>> coalib.bearlib.aspects['shortlog.colonExistence']
    <aspectclass 'Root.Metadata.CommitMessage.Shortlog.ColonExistence'>
    """

    def __init__(self, module):
        """
        Take over all members from original `module` object and load all
        direct ``Root.subaspect`` classes from their corresponding submodules
        (all submodules starting with uppercase letters), and add them as
        members to this wrapper module.
        """
        super().__init__(module.__name__)
        self.__dict__.update(getmembers(module))
        for _, submodname, _ in iter_modules(module.__path__):
            if submodname[0].isupper():
                submod = import_module('.' + submodname, module.__name__)
                subaspect = getattr(submod, submodname)
                setattr(self, submodname, subaspect)

    def __getitem__(self, aspectname):
        regex = re.compile('(^|\.)%s$' % aspectname.lower())
        matches = []

        def search(aspects):
            """
            Recursively search in `aspects` for those matching
            case-insensitively the given ``aspectname``.
            """
            for aspect in aspects:
                if regex.search(aspect.__qualname__.lower()):
                    matches.append(aspect)
                if aspect.subaspects:
                    search(aspect.subaspects.values())

        search([Root])
        if not matches:
            raise LookupError('no aspect named %s' % repr(aspectname))

        if len(matches) > 1:
            raise LookupError('multiple aspects named %s. choose from %s' % (
                repr(aspectname),
                repr(sorted(matches, key=lambda a: a.__qualname__))))

        return matches[0]


# replace original module with wrapper
sys.modules[__name__] = aspectsModule(sys.modules[__name__])



import traceback
import logging

from coalib.output.printers.LOG_LEVEL import LOG_LEVEL
from coalib.processes.communication.LogMessage import LogMessage


class LogPrinterMixin:
    """
    Provides access to the logging interfaces (e.g. err, warn, info) by routing
    them to the log_message method, which should be implemented by descendants
    of this class.
    """

    def debug(self, *messages, delimiter=' ', timestamp=None, **kwargs):
        self.log_message(LogMessage(LOG_LEVEL.DEBUG,
                                    *messages,
                                    delimiter=delimiter,
                                    timestamp=timestamp),
                         **kwargs)

    def info(self, *messages, delimiter=' ', timestamp=None, **kwargs):
        self.log_message(LogMessage(LOG_LEVEL.INFO,
                                    *messages,
                                    delimiter=delimiter,
                                    timestamp=timestamp),
                         **kwargs)

    def warn(self, *messages, delimiter=' ', timestamp=None, **kwargs):
        self.log_message(LogMessage(LOG_LEVEL.WARNING,
                                    *messages,
                                    delimiter=delimiter,
                                    timestamp=timestamp),
                         **kwargs)

    def err(self, *messages, delimiter=' ', timestamp=None, **kwargs):
        self.log_message(LogMessage(LOG_LEVEL.ERROR,
                                    *messages,
                                    delimiter=delimiter,
                                    timestamp=timestamp),
                         **kwargs)

    def log(self, log_level, message, timestamp=None, **kwargs):
        self.log_message(LogMessage(log_level,
                                    message,
                                    timestamp=timestamp),
                         **kwargs)

    def log_exception(self,
                      message,
                      exception,
                      log_level=LOG_LEVEL.ERROR,
                      timestamp=None,
                      **kwargs):
        """
        If the log_level of the printer is greater than DEBUG, it prints
        only the message. If it is DEBUG or lower, it shows the message
        along with the traceback of the exception.

        :param message:   The message to print.
        :param exception: The exception to print.
        :param log_level: The log_level of this message (not used when
                          logging the traceback. Tracebacks always have
                          a level of DEBUG).
        :param timestamp: The time at which this log occurred. Defaults to
                          the current time.
        :param kwargs:    Keyword arguments to be passed when logging the
                          message (not used when logging the traceback).
        """
        if not isinstance(exception, BaseException):
            raise TypeError('log_exception can only log derivatives of '
                            'BaseException.')

        traceback_str = '\n'.join(
            traceback.format_exception(type(exception),
                                       exception,
                                       exception.__traceback__))

        self.log(log_level, message, timestamp=timestamp, **kwargs)
        self.log_message(
            LogMessage(LOG_LEVEL.INFO,
                       'Exception was:' + '\n' + traceback_str,
                       timestamp=timestamp),
            **kwargs)

    def log_message(self, log_message, **kwargs):
        """
        It is your reponsibility to implement this method, if you're using this
        mixin.
        """
        raise NotImplementedError


class LogPrinter(LogPrinterMixin):
    """
    This class is deprecated and will be soon removed. To get logger use
    logging.getLogger(__name__). Make sure that you're getting it when the
    logging configuration is loaded.

    The LogPrinter class allows to print log messages to an underlying Printer.

    This class is an adapter, means you can create a LogPrinter from every
    existing Printer instance.
    """

    def __init__(self,
                 printer=None,
                 log_level=LOG_LEVEL.DEBUG,
                 timestamp_format='%X'):
        """
        Creates a new log printer from an existing Printer.

        :param printer:          The underlying Printer where log messages
                                 shall be written to. If you inherit from
                                 LogPrinter, set it to self.
        :param log_level:        The minimum log level, everything below will
                                 not be logged.
        :param timestamp_format: The format string for the
                                 datetime.today().strftime(format) method.
        """
        self.logger = logging.getLogger()

        self._printer = printer
        self.log_level = log_level
        self.timestamp_format = timestamp_format

    @property
    def log_level(self):
        """
        Returns current log_level used in logger.
        """
        return self.logger.getEffectiveLevel()

    @log_level.setter
    def log_level(self, log_level):
        """
        Sets log_level for logger.
        """
        self.logger.setLevel(log_level)

    @property
    def printer(self):
        """
        Returns the underlying printer where logs are printed to.
        """
        return self._printer

    def log_message(self, log_message, **kwargs):
        if not isinstance(log_message, LogMessage):
            raise TypeError('log_message should be of type LogMessage.')
        self.logger.log(log_message.log_level, log_message.message)

    def __getstate__(self):
        # on Windows there are problems with serializing loggers, so omit it
        oldict = self.__dict__.copy()
        del oldict['logger']
        return oldict

    def __setstate__(self, newdict):
        self.__dict__.update(newdict)
        # restore logger by name
        self.logger = logging.getLogger()



from coalib.bearlib.aspects import Root


@Root.subaspect
class Spelling:
    """
    This aspect describes spelling of your source code.
    """


@Spelling.subaspect
class aspectsYEAH:
    """
    This aspect dictates that the term ``aspects`` and ``aspect`` must have
    all letters in lower case and the term ``aspectsYEAH`` must match the
    exact expression.
    """
    class docs:
        example = """
        Valid Cases:
        # aspects are the mother of all futures of coala
        # ``aspectsYEAH`` project is simply awesome
        Invalid Cases:
        # Aspects are the mother of all futures of coala
        # ``aspectsyeah`` project is simply awesome
        """
        example_language = 'All'
        importance_reason = """
        The concepts of aspects in coala are too fundamental to contain any
        upper-case letters.
        """
        fix_suggestions = """
        ``aspects`` or ``aspect`` are always written with all lower-case
        letters or the term ``aspectsYEAH`` must match the exact expression.
        """
